﻿Imports System.IO
Imports System.Text
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media


Public Class PageProvider
    Implements IPageProvider
    Private Const FirstLineIndent As Integer = 3 ' in units of space width
    Private Const ParagraphSpacing As Double = 1.0 / 3 ' in units of character height

    Private fontMetrics As New FontMetrics(New Font())
    Private fontSize As Double
    Private pageSize As Size

    Public Sub New(ByVal book As BookInfo)
        Me.Book = book
    End Sub

    ' Only public properties
    Private _book As BookInfo
    Public Property Book() As BookInfo
        Get
            Return _book
        End Get
        Protected Set(ByVal value As BookInfo)
            _book = value
        End Set
    End Property

    Public Function SetPageSize(ByVal size As Size) As Boolean Implements IPageProvider.SetPageSize
        Dim invalidate As Boolean = False
        Me.pageSize = size

        If Me.Book.PaginationPageSize <> pageSize Then
            InvalidatePaginationInformation()
            Me.Book.PaginationPageSize = pageSize
            invalidate = True
        End If
        Return invalidate
    End Function

    Public Function SetFont(ByVal fontFamily As FontFamily, ByVal fontSize As Double) As Boolean Implements IPageProvider.SetFont
        If fontFamily IsNot fontMetrics.Font.FontFamily Then
            Me.fontMetrics = New FontMetrics(New Font(fontFamily.Source))
        End If

        Dim invalidate As Boolean = False
        Me.fontSize = fontSize

        If Me.Book.PaginationFontFamily <> fontFamily.Source OrElse Me.Book.PaginationFontSize <> fontSize Then
            InvalidatePaginationInformation()
            Me.Book.PaginationFontFamily = fontFamily.Source
            Me.Book.PaginationFontSize = fontSize
            invalidate = True
        End If
        Return invalidate
    End Function

    Private Sub InvalidatePaginationInformation()
        For Each chapter In Me.Book.Chapters
            Dim pageCount As Integer = chapter.Pages.Count

            If pageCount > 1 Then
                chapter.Pages.RemoveRange(1, pageCount - 1)
            End If

            chapter.Pages(0).IsLastPageInChapter = False
            chapter.Pages(0).IsPaginated = False
            chapter.Pages(0).Words.Clear()
        Next chapter

        Me.Book.IsChapterAndPageInvalid = True
    End Sub

    Public Function GetPage(ByVal characterCount As Integer, <System.Runtime.InteropServices.Out()> ByRef chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetPage
        chapterIndex = 0
        pageIndex = 0

        For ch = 0 To Me.Book.Chapters.Count - 1
            If Me.Book.Chapters(ch).Pages(0).AccumulatedCharacterCount <= characterCount Then
                chapterIndex = ch
            Else
                Exit For
            End If
        Next ch

        Dim pg As Integer = 0
        Do
            Dim success As Boolean = Paginate(chapterIndex, pg)

            If success AndAlso Me.Book.Chapters(chapterIndex).Pages(pg).AccumulatedCharacterCount <= characterCount Then
                pageIndex = pg
            Else
                Exit Do
            End If
            pg += 1
        Loop
        Return GetPage(chapterIndex, pageIndex)
    End Function

    Public Function GetPage(ByVal chapterIndex As Integer, ByVal pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetPage
        If pageSize.IsEmpty Then
            Return Nothing
        End If

        If pageIndex < 0 Then
            Return Nothing
        End If

        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            Return Nothing
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

        ' Check if the last page of the paragraph has already been established, 
        '  and the requested page is greater than that last page
        If chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter AndAlso pageIndex > chapter.Pages.Count - 1 Then
            Return Nothing
        End If

        ' If the last page of a chapter is requested, and it's already available, get it
        If pageIndex < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsLastPageInChapter Then
            Return BuildPageElement(chapter, chapter.Pages(pageIndex))
        End If

        ' Otherwise, if the page has already been paginated, get it
        If pageIndex + 1 < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsPaginated Then
            Return BuildPageElement(chapter, chapter.Pages(pageIndex))
        End If

        If pageIndex <> chapter.Pages.Count - 1 Then
            Throw New Exception(String.Format("GetPage can only paginate one page per call: pageIndex = {0} while chapter.Pages.Count = {1}", pageIndex, chapter.Pages.Count))
        End If

        Dim pageInfo As PageInfo = chapter.Pages(pageIndex)

        Dim nextPageInfo As PageInfo = Paginate(chapter, pageInfo)

        If nextPageInfo IsNot Nothing Then ' ie, if !pageInfo.IsLastPageInChapter
            chapter.Pages.Add(nextPageInfo)
        End If

        Return BuildPageElement(chapter, pageInfo)
    End Function

    Public Function QueryLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.QueryLastPage
        Return GetLastPage(chapterIndex, pageIndex, False)
    End Function

    Public Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer) As FrameworkElement Implements IPageProvider.GetLastPage
        Dim element As FrameworkElement = GetLastPage(chapterIndex, pageIndex, True)
        Return element
    End Function

    Private Function GetLastPage(ByVal chapterIndex As Integer, <System.Runtime.InteropServices.Out()> ByRef pageIndex As Integer, ByVal paginateIfNecessary As Boolean) As FrameworkElement
        If pageSize.IsEmpty Then
            pageIndex = -1
            Return Nothing
        End If

        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            pageIndex = -1
            Return Nothing
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)
        Dim lastIndex As Integer = chapter.Pages.Count - 1

        If chapter.Pages(lastIndex).IsLastPageInChapter AndAlso chapter.Pages(lastIndex).IsPaginated Then ' IsPaginated should be implied
            pageIndex = lastIndex
            Return BuildPageElement(chapter, chapter.Pages(lastIndex))
        End If

        If Not paginateIfNecessary Then
            pageIndex = -1
            Return Nothing
        End If

        Do While Not chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter
            Paginate(chapterIndex, chapter.Pages.Count - 1)
        Loop

        pageIndex = chapter.Pages.Count - 1
        Return BuildPageElement(chapter, chapter.Pages(chapter.Pages.Count - 1))
    End Function

    Private Function Paginate(ByVal chapterIndex As Integer, ByVal pageIndex As Integer) As Boolean
        If chapterIndex < 0 OrElse chapterIndex > Me.Book.Chapters.Count - 1 Then
            Return False
        End If

        Dim chapter As ChapterInfo = Me.Book.Chapters(chapterIndex)

        If pageIndex < chapter.Pages.Count AndAlso chapter.Pages(pageIndex).IsPaginated Then
            Return True
        End If

        If chapter.Pages(chapter.Pages.Count - 1).IsLastPageInChapter Then
            Return False
        End If

        Dim pageInfo As PageInfo = chapter.Pages(pageIndex)
        Dim nextPageInfo As PageInfo = Paginate(chapter, pageInfo)

        If nextPageInfo IsNot Nothing Then
            chapter.Pages.Add(nextPageInfo)
        End If

        Return True
    End Function

    ' TODO: Could pass chapter and pageIndex here, to obtain pageInfo within, and then can set nextPageInfo
    Private Function Paginate(ByVal chapter As ChapterInfo, ByVal pageInfo As PageInfo) As PageInfo
        ' Get dimensions of space character
        Dim spaceSize As Size = fontMetrics.MeasureText(" ")
        Dim lineHeight As Integer = CInt(Fix(Math.Ceiling(fontSize * spaceSize.Height)))
        Dim spaceWidth As Integer = CInt(Fix(Math.Ceiling(fontSize * spaceSize.Width)))

        ' Initialize a couple variables
        Dim accumHeight As Integer = 0
        Dim characterCount As Integer = pageInfo.AccumulatedCharacterCount
        Dim nextPageInfo As PageInfo = Nothing
        Dim paragraphIndex As Integer = pageInfo.ParagraphIndex
        Dim characterIndex As Integer = pageInfo.CharacterIndex

        ' Loop for lines on page
        Do
            Dim paragraph As ParagraphInfo = Me.Book.Paragraphs(paragraphIndex)

            ' For blank paragraphs, leave a gap...
            If paragraph.Text.Length = 0 Then
                ' .. but not when it's at the top of the page
                If accumHeight > 0 Then
                    accumHeight += CInt(Fix(Math.Ceiling(ParagraphSpacing * lineHeight)))
                End If
                ' Normal case for non-blank paragraphs
            Else
                ' This is the amount to indent the entire paragraph
                Dim indent As Integer = spaceWidth * paragraph.Indent

                ' Apply first line indent if paragraph indent is zero, and it's
                '  not the first paragraph in the chapter
                If indent = 0 AndAlso characterIndex = 0 AndAlso paragraph IsNot Me.Book.Paragraphs(chapter.FirstParagraph) Then
                    indent = spaceWidth * FirstLineIndent
                End If

                ' Start off accumulated line width with the indent
                Dim accumWidth As Integer = indent

                ' Skip blanks that might be at the beginning of a line.
                '  (This can happen if sentences are separated by two
                '      blanks, and a sentence ends in the preceding line.)
                Do While paragraph.Text.Chars(characterIndex) = " "c
                    characterIndex += 1
                Loop

                ' Loop for words within line -- always the same paragraph
                Do
                    ' Find next space in paragraph
                    Dim spaceIndex As Integer = paragraph.Text.IndexOf(" "c, characterIndex)
                    Dim wordCharacters As Integer = (If(spaceIndex = -1, paragraph.Text.Length, spaceIndex)) - characterIndex
                    Dim wordWidth As Integer = CInt(Fix(Math.Ceiling(fontSize * fontMetrics.MeasureText(paragraph.Text, characterIndex, wordCharacters).Width)))

                    ' Check if the accumulated width exceeds the page width, 
                    '      but make sure to have at least one word.
                    If accumWidth > indent AndAlso accumWidth + wordWidth >= pageSize.Width Then
                        Exit Do
                    End If

                    Dim word As WordInfo = New WordInfo With {.LocationLeft = accumWidth, .LocationTop = accumHeight, .ParagraphIndex = paragraphIndex, .CharacterIndex = characterIndex, .CharacterCount = wordCharacters}
                    pageInfo.Words.Add(word)

                    accumWidth += wordWidth
                    characterCount += wordCharacters + (If(spaceIndex = -1, 0, 1))
                    characterIndex += wordCharacters + (If(spaceIndex = -1, 0, 1))

                    ' End of line because it's the end of the paragraph
                    If spaceIndex = -1 Then
                        characterIndex = 0
                        Exit Do
                    End If

                    ' End of line because no more words will fit
                    If accumWidth + spaceWidth >= pageSize.Width Then
                        Exit Do
                    End If

                    accumWidth += spaceWidth

                    ' Otherwise, point to next word
                    characterIndex = spaceIndex + 1

                    ' ready for next word
                Loop
                ' Bump up the accumulated height of the page
                accumHeight += lineHeight
            End If

            ' Start a new line, but check first if it's a new paragraph
            If characterIndex = 0 Then
                ' If we've been working with the last paragraph in the chapter,
                '      we're also done with the page
                If paragraphIndex + 1 = chapter.FirstParagraph + chapter.ParagraphCount Then
                    pageInfo.IsLastPageInChapter = True
                    Exit Do
                End If
                ' Otherwise, kick up the paragraphIndex
                paragraphIndex += 1
            End If

            ' Check if the next paragraph is too much for this page
            If (Me.Book.Paragraphs(paragraphIndex).Text.Length = 0 AndAlso accumHeight + lineHeight \ 2 > pageSize.Height) OrElse accumHeight + lineHeight > pageSize.Height Then
                Exit Do
            End If
        Loop

        pageInfo.IsPaginated = True

        ' Create PageInfo for next page
        If Not pageInfo.IsLastPageInChapter Then
            nextPageInfo = New PageInfo With {.ParagraphIndex = paragraphIndex, .CharacterIndex = characterIndex, .AccumulatedCharacterCount = characterCount}
        End If

        Return nextPageInfo
    End Function

    Private Function BuildPageElement(ByVal chapter As ChapterInfo, ByVal pageInfo As PageInfo) As FrameworkElement
        If pageInfo.Words.Count = 0 Then
            Paginate(chapter, pageInfo)
        End If

        Dim canvas As New Canvas()

        For Each word In pageInfo.Words
            Dim txtblk As TextBlock = New TextBlock With {.FontFamily = fontMetrics.Font.FontFamily, .FontSize = Me.fontSize, .Text = Me.Book.Paragraphs(word.ParagraphIndex).Text.Substring(word.CharacterIndex, word.CharacterCount), .Tag = word}

            canvas.SetLeft(txtblk, word.LocationLeft)
            canvas.SetTop(txtblk, word.LocationTop)
            canvas.Children.Add(txtblk)
        Next word

        ' Avoid heavy build-up of WordInfo collections
        pageInfo.Words.Clear()

        Return canvas
    End Function

    Public Function FindText(ByVal paragraphIndex As Integer, ByVal characterIndex As Integer, ByVal text As String, ByVal matchCase As Boolean, ByVal wholeWord As Boolean, ByVal goUp As Boolean) As FindInfo
        If paragraphIndex < 0 OrElse paragraphIndex >= Me.Book.Paragraphs.Count Then
            Return Nothing
        End If

        Dim comparisonType As StringComparison = If(matchCase, StringComparison.Ordinal, StringComparison.OrdinalIgnoreCase)

        If Not goUp Then
            Return FindTextDown(paragraphIndex, characterIndex, text, wholeWord, comparisonType)
        Else
            Return FindTextUp(paragraphIndex, characterIndex, text, wholeWord, comparisonType)
        End If
    End Function

    Private Function FindTextDown(ByVal paragraphIndex As Integer, ByVal characterIndex As Integer, ByVal text As String, ByVal wholeWord As Boolean, ByVal comparisonType As StringComparison) As FindInfo
        Do While paragraphIndex < Me.Book.Paragraphs.Count
            Dim paragraphText As String = Me.Book.Paragraphs(paragraphIndex).Text

            Do While characterIndex + text.Length <= paragraphText.Length
                Dim findIndex As Integer = paragraphText.IndexOf(text, characterIndex, comparisonType)

                If findIndex = -1 Then
                    Exit Do
                ElseIf (Not wholeWord) OrElse TestWholeWord(paragraphText, findIndex, text.Length) Then
                    Return CreateFindInfo(paragraphIndex, findIndex, text.Length)
                Else
                    characterIndex = findIndex + 1
                End If
            Loop

            ' Go to the next paragraph
            paragraphIndex += 1
            characterIndex = 0
        Loop
        Return Nothing
    End Function

    Private Function FindTextUp(ByVal paragraphIndex As Integer, ByVal characterIndex As Integer, ByVal text As String, ByVal wholeWord As Boolean, ByVal comparisonType As StringComparison) As FindInfo
        characterIndex = Math.Min(Me.Book.Paragraphs(paragraphIndex).Text.Length - 1, characterIndex + text.Length - 2)

        Do While paragraphIndex >= 0
            Dim paragraphText As String = Me.Book.Paragraphs(paragraphIndex).Text

            Do While characterIndex > text.Length - 2
                Dim findIndex As Integer = paragraphText.LastIndexOf(text, characterIndex, comparisonType)

                If findIndex = -1 Then
                    Exit Do
                ElseIf (Not wholeWord) OrElse TestWholeWord(paragraphText, findIndex, text.Length) Then
                    Return CreateFindInfo(paragraphIndex, findIndex, text.Length)
                Else
                    characterIndex = findIndex - 1
                End If
            Loop

            ' Go to the previous paragraph
            paragraphIndex -= 1

            If paragraphIndex >= 0 Then
                characterIndex = Me.Book.Paragraphs(paragraphIndex).Text.Length - 1
            End If
        Loop
        Return Nothing
    End Function

    Private Function TestWholeWord(ByVal paragraphText As String, ByVal findIndex As Integer, ByVal findTextLength As Integer) As Boolean
        ' If wholeWord is true, then the matched text must be either 
        '      at the beginning of the paragraph or be prefaced
        '      by a space or punctuation symbol AND
        '      it must either be at the end of the paragraph or be
        '      followed by a space or punctuation.
        Dim prevCharIndex As Integer = findIndex - 1
        Dim nextCharIndex As Integer = findIndex + findTextLength
        Dim isMatch As Boolean = (prevCharIndex = -1 OrElse paragraphText.Chars(prevCharIndex) = " "c OrElse Char.IsPunctuation(paragraphText.Chars(prevCharIndex))) AndAlso (nextCharIndex = paragraphText.Length OrElse paragraphText.Chars(nextCharIndex) = " "c OrElse Char.IsPunctuation(paragraphText.Chars(nextCharIndex)))
        Return isMatch
    End Function

    Private Function CreateFindInfo(ByVal paragraphIndex As Integer, ByVal characterIndex As Integer, ByVal findTextLength As Integer) As FindInfo
        Dim characterOffset As Integer = Me.Book.Paragraphs(paragraphIndex).AccumulatedCharacterCount + characterIndex

        Dim findInfo As FindInfo = New FindInfo With {.CharacterOffset = characterOffset, .ParagraphIndex = paragraphIndex, .StartCharacterIndex = characterIndex, .EndCharacterIndex = characterIndex + findTextLength}

        Return findInfo
    End Function
End Class


